/*
 Calisto (c) 1998-1999 Peter Howkins, Matthew Howkins, Simon Howkins

 $Id: commands.c,v 1.18 2000/01/08 22:52:04 peter Exp $

 $Log: commands.c,v $
 Revision 1.18  2000/01/08 22:52:04  peter
 Changed commands to use help_usage() to print usage instructions

 Revision 1.17  2000/01/06 21:23:00  peter
 added command remote, changed output of tell slightly

 Revision 1.16  2000/01/06 21:02:58  peter
 modified finger's handling of privs, adding colour

 Revision 1.15  2000/01/02 16:18:21  peter
 Added command_sumotd to go with the sumotd printed during login

 Revision 1.14  2000/01/02 16:08:55  peter
 Updated for new return value of load_player

 Revision 1.13  2000/01/02 15:36:47  peter
 Changed PRIV_AD_GENERAL to PRIV_AD_BASE

 Revision 1.12  1999/12/28 20:42:30  peter
 changed the logout messages to be more specific

 Revision 1.11  1999/12/28 19:51:17  peter
 Changed command 'netstat' to 'stat net' and added 'stat mem'

 Revision 1.10  1999/12/27 00:31:14  peter
 #includes "config.h"

 Revision 1.9  1999/12/27 00:28:38  peter
 Improved command_group() and added a basic command_remove()

 Revision 1.8  1999/12/20 20:53:12  peter
 Improved command_grant()

 Revision 1.7  1999/12/19 16:01:25  peter
 Added cls command, and synonym clear

 Revision 1.6  1999/12/18 18:09:04  peter
 Added command_reload() for reloading text files.
 Currently only uses help_reload()

 Revision 1.5  1999/12/15 23:27:58  peter
 Added command_help() found in help.c

 Revision 1.4  1999/12/14 20:50:38  peter
 Added basic netstat command

 Revision 1.3  1999/12/11 23:10:18  peter
 *** empty log message ***

 Revision 1.2  1999/12/03 22:30:18  peter
 added term width + height to finger

 Revision 1.1  1999/11/23 22:26:54  peter
 Initial revision


 */
static char rcsid[] = "$Id: commands.c,v 1.18 2000/01/08 22:52:04 peter Exp $";

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "unistd.h"

#include "config.h"
#include "colours.h"
#include "commands.h"
#include "globals.h"
#include "help.h"
#include "library.h"
#include "log.h"
#include "playerdb.h"
#include "privs.h"
#include "structs.h"
#include "strplus.h"

typedef struct {
  const char *name;
  const char *shortcut;
  void (*function)(character *, char *);
  unsigned long priv;
} command;

void command_bug(character *, char *);
void command_cls(character *, char *);
void command_commands(character *, char *);
void command_dc(character *, char *);
void command_emote(character *, char *);
void command_finger(character *, char *);
void command_grant(character *, char *);
void command_group(character *, char *);
void command_idea(character *, char *);
void command_lsd(character *, char *);
void command_motd(character *, char *);
void command_ping(character *, char *);
void command_prompt(character *, char *);
void command_quit(character *, char *);
void command_reload(character *, char *);
void command_remote(character *, char *);
void command_remove(character *, char *);
void command_say(character *, char *);
void command_shout(character *, char *);
void command_shutdown(character *, char *);
void command_stat(character *, char *);
void command_sumotd(character *, char *);
void command_tell(character *, char *);
void command_time(character *, char *);
void command_typo(character *, char *);
void command_users(character *, char *);
void command_version(character *, char *);
void command_who(character *, char *);
void command_whoami(character *, char *);

const command commands[] = {
  { "bug",     "",  command_bug,      PRIV_BASE },
  { "bump",    "",  command_dc,       PRIV_SU_BUMP },
  { "clear",   "",  command_cls,      PRIV_BASE },
  { "cls",     "",  command_cls,      PRIV_BASE },
  { "commands","",  command_commands, PRIV_BASE },
  { "dc",      "",  command_dc,       PRIV_SU_BUMP },
  { "emote",   ";:",command_emote,    PRIV_TALK },
  { "exit",    "",  command_quit,     PRIV_BASE },
  { "finger",  "",  command_finger,   PRIV_BASE },
  { "grant",   "",  command_grant,    PRIV_AD_GRANT },
  { "group",   "",  command_group,    PRIV_BASE },
  { "help",    "",  command_help,     PRIV_BASE }, /* Found in help.c */
  { "idea",    "",  command_idea,     PRIV_BASE },
  { "lsd",     "",  command_lsd,      PRIV_SU_TRACE },
  { "motd",    "",  command_motd,     PRIV_BASE },
  { "ping",    "",  command_ping,     PRIV_BASE },
  { "prompt",  "",  command_prompt,   PRIV_BASE },
  { "quit",    "",  command_quit,     PRIV_BASE },
  { "reload",  "",  command_reload,   PRIV_AD_BASE },
  { "remote",  ",", command_remote,   PRIV_TALK },
  { "remove",  "",  command_remove,   PRIV_AD_GRANT },
  { "say",     "'", command_say,      PRIV_TALK },
  { "shout",   "",  command_shout,    PRIV_TALK },
  { "shutdown","",  command_shutdown, PRIV_AD_SHUTDOWN },
  { "stat",    "",  command_stat,     PRIV_AD_BASE },
  { "sumotd",  "",  command_sumotd,   PRIV_SU_BASE },
  { "tell",    ".", command_tell,     PRIV_TALK },
  { "time",    "",  command_time,     PRIV_BASE },
  { "typo",    "",  command_typo,     PRIV_BASE },
  { "users",   "",  command_users,    PRIV_BASE },
  { "version", "",  command_version,  PRIV_BASE },
  { "who",     "",  command_who,      PRIV_BASE },
  { "whoami",  "",  command_whoami,   PRIV_BASE },
  { "x",       "",  command_finger,   PRIV_BASE }
};
const unsigned commandn = sizeof commands / sizeof(command);

void command_version(character *c, char *r)
{
  descriptor *des = getdes(c);

  bar2(des, "Version", "^Y", "^R");
  send_to_char(c, "Calisto %u.%02u (c) 1998-2000 Flibble, The_Pope and Wilfred\n", version/100, version%100);
  send_to_char(c, "  ^G+^n Colour Code ^G-^n ^YCirrus^n\n");
  bar2(des, NULL, NULL, "^R");
}

void command_sumotd(character *c, char *r)
{
  descriptor *des = getdes(c);

  bar2(des, "Super-user Message Of The Day", "^Y", "^R");
  send_file_to_descriptor("lib/etc/sumotd", des);
  bar2(des, NULL, NULL, "^R");
}

void command_cls(character *c, char *r)
{
  descriptor *des = getdes(c);
  int i;

  for (i = 0; i < des->term_height; i++) {
    send_to_char(c, "\n");
  }
  send_to_char(c, "\x1b[0;0H");
}

void command_reload(character *c, char *r)
{
  if (STRIEQ(r, "help")) {
    help_reload(c);
  } else if (STRIEQ(r, "all")) {
    help_reload(c);
  } else {
    send_to_char(c, "Unrecognised reload option %s\n", r);
  }
}

#define COMMAND_STAT_OPTIONS "net mem"
static void command_stat_usage(character *c)
{
  send_to_char(c, "Usage: stat <option> [<option>...]\n");
  send_to_char(c, "where <option> is one of `%s all`\n",
               COMMAND_STAT_OPTIONS);
}

void command_stat(character *c, char *r)
{
  if (*r) {
    while (*r) {
      char *sub;

      strmunch(r, &sub, &r);
      if (STRIEQ(sub, "mem")) {
        pool_stat_t ps = pool_stat(descriptor_pool);

        send_to_char(c, "^H[Memory - Descriptors]^n\n");
        send_to_char(c, "  Block Size : %u\n", ps.block_size);
        send_to_char(c, "       Usage : %u\n", ps.usage);
        send_to_char(c, "    Capacity : %u (%.1fk)\n", ps.capacity,
                     (double) (ps.block_size * ps.capacity) / 1024.0);
        send_to_char(c, "        Free : %u\n", ps.capacity - ps.usage);
      } else if (STRIEQ(sub, "net")) {
        send_to_char(c, "^H[Network and Machine]^n\n");
        send_to_char(c, "   PID : %u\n", (unsigned) getpid());
        send_to_char(c, "  PPID : %u\n", (unsigned) getppid());
        send_to_char(c, "  HOST : %s\n", getenv("HOSTNAME"));
        send_to_char(c, "  TYPE : %s\n", getenv("MACHTYPE"));
      } else if (STRIEQ(sub, "all")) {
        char all[] = COMMAND_STAT_OPTIONS;

        command_stat(c, all);
      } else {
        send_to_char(c, "Unknown stat option `%s`\n", sub);
        command_stat_usage(c);
      }
    }
  } else {
    command_stat_usage(c);
  }
#undef COMMAND_STAT_OPTIONS
}

void command_finger(character *c, char *r)
{
  /* formats 
     finger       - does a finger on the character that called func
     finger bob   - does a finger on bob
     check for
       r = ""
       r logged in 
       r not logged in and exists
       r not logged in and doesn't exist
  */
  time_t now;
  descriptor *des = getdes(c);
  descriptor *victdes;
  character *victim;
  character target;
  char buffer[OUTPUT_BUFFER];
  bool loggedin = FALSE;

  if(STREQ(r, "")) {
    victim = c;
    loggedin = TRUE;
  } else {
    if((victim = character_from_name(r)) != NULL) {
      /* victim is logged on and found */
      loggedin = TRUE;
    } else {
      /* check to see is r is in the players directory, if so load
         else return */
      int failed;
      failed = load_player(&target, r);
      if (failed) {
        send_to_char(c, "Player %s, not found\n", r);
        return;
      } else {
        victim = &target;
        loggedin = FALSE;
      }
    }
  }

  now = time(NULL);
  victdes = getdes(victim);
  if (loggedin) {
    seconds_to_string(victim->prevtime + difftime(now, victim->logintime),
                    buffer);
  } else {
    seconds_to_string(victim->prevtime, buffer);
  }
  bar2(des, "Finger", NULL, NULL);
  send_to_char(c, "      Name: %s\n", victim->name);
  send_to_char(c, "Total Time: %s\n", buffer);
  /* Display privileges */
  if (victim == c || haspriv(c, PRIV_SU_BASE)) {
    unsigned i;
    send_to_char(c, "     Privs:");
    for (i = 0; i < privn; i++)
      if (haspriv(victim, privs[i].priv))
        send_to_char(c, " %s%s", privcol(privs[i].priv), privs[i].privname);
    send_to_char(c, "^n %08lx\n", victim->privs);
  }
  if (loggedin) {
    send_to_char(c, "     Group: %s\n", victim->group);
    send_to_char(c, "  Termtype: %s (%ux%u)\n", victdes->termtype,
                 victdes->term_width, victdes->term_height);
    if (haspriv(c, PRIV_SU_TRACE)) {
      send_to_char(c, "      Host: %s\n", victdes->hostname);
    }
  }
  bar2(des, NULL, NULL, NULL);
}

void command_group(character *c, char *r)
{
  char groupname[MAX_GROUP_NAME_LENGTH+1];

  stripCodes(groupname, r);

  if(STREQ(r, "")) {
    strcpy(groupname, "Public");
  }
  if (strlen(r) > MAX_GROUP_NAME_LENGTH) {
    send_to_char(c, "Group name too long (max %d)\n", MAX_GROUP_NAME_LENGTH);
    return;
  }

  strlower(groupname);
  groupname[0] = toupper(groupname[0]);

  if (!STREQ(groupname, c->group)) {
    /* Announce departure of character from old group */
    send_to_room_except(c->group, c, "+++ %s joins group %s\n",
                        c->name, groupname);
    strcpy(c->group, groupname);
    /* annouce arrival in new group */
    send_to_char(c, "+++ You have joined group ^y%s^n\n", c->group);
    send_to_room_except(c->group, c, "+++ %s joins this group\n", c->name);
  } else {
    send_to_char(c, "You are already in group ^y%s^n\n", c->group);
  }
}

void command_dc(character *c, char *r)
{
  descriptor *des = NULL;
  character *ch;

  if (STREQ(r, "")) {
    help_usage(c, "bump/dc");
    return;
  }

  ch = character_from_name(r);
  if (ch) {
    /* found character, drop em */
    des = getdes(ch);
  } else {
    /* char name not found - assume a number */
    if (!strisdigit(r)) {
      /* Not a number - report bad character name */
      send_to_char(c, "Character %s not known\n", r);
      return;
    } else {
      int sfd = atoi(r);
      listnode *dNode = AllConns.head.next;

      while (LIST_NODE_IS_REAL(dNode)) {
        descriptor *des2 = LIST_GET_DATA(dNode, descriptor *, descriptorlink);
        if (sfd == des2->sfd) {
          des = des2;
          break;
        }
        dNode = dNode->next;
      }
      if (!des) { 
        send_to_char(c, "SFD %d not found\n", sfd);
        return;
      }
    }
  }

  /* Before we drop, check we are not dropping ourselves */
  if (des == getdes(c)) {
    send_to_char(c, "Haven't you heard of the quit command?\n");
    return;
  }
  des->state = STATE_CLOSING;
  if (des->player.loggedin) {
    send_to_all_except(&des->player, "\n+++ ^r%s^n has logged out\n",
                       des->player.name);
  }
}

void command_time(character *c, char *r)
{
  time_t now;
  struct tm nowtm;
  char buff[OUTPUT_BUFFER];
  unsigned long howlong;

  now = time(NULL);
  nowtm = *localtime(&now);
  strftime(buff, OUTPUT_BUFFER, "%a %b %d %H:%M:%S %Y", &nowtm);
  send_to_char(c, "^WCurrent  : %s\n^n", buff);

  nowtm = *localtime(&starttime);
  strftime(buff, OUTPUT_BUFFER, "%a %b %d %H:%M:%S %Y", &nowtm);
  send_to_char(c, "Up Since : %s\n", buff);

  howlong = (unsigned long) difftime(now, starttime);
  seconds_to_string(howlong, buff);
  send_to_char(c, "Up Time  : %s\n", buff);
}

void command_ping(character *c, char *r) 
{
  character *target;

  if (STREQ(r, "")) {
    help_usage(c, "ping");
  } else {  
    target = character_from_name(r);
    if (target == NULL) {
      send_to_char(c, "%s not found\n", r);
    } else {
      send_to_char(target, "\a^YPING\n^n%s is trying to contact you\n",
                   c->name);
    }
  }
}

void command_remove(character *c, char *r)
{
  if (STREQ(r, "")) {
    help_usage(c, "remove");
    return;
  }
  {
    character *target;
    char *privname = r;
    const priv *p;
   
    strmunch(r, &r, &privname);
    /* r should now be player name, privname should be the priv */
    /* try to find player, if not found bail out of function */
    target = character_from_name(r);
    if (!target) {
      send_to_char(c, "Player %s not found\n", r);
      help_usage(c, "remove");
      return;
    }
    /* check if removing privs from oneself */
    if (c == target) {
      send_to_char(c, "You may not remove your own privs\n");
      return;
    }
    /* if privname is blank print out usage and bail out */
    if (STREQ(privname, "")) {
      send_to_char(c, "Unknown priv\n");
      help_usage(c, "remove");
      return;
    }
    /* search through the priv list and try to find the priv */
    p = privname_to_priv(privname);
    if (p) {
      send_to_char(c, "You have removed from %s the %s priv.\n",
                   target->name, p->privname);
      send_to_char(target, "You have lost the %s priv\n",
                   p->privname);
      clearpriv(target, p->priv);
    } else {
      send_to_char(c, "Priv %s not found\n", privname);
      help_usage(c, "remove");
    }
  }
}

static void command_grant_usage(character *c)
{
  unsigned i;

  help_usage(c, "grant");
  send_to_char(c, "Available privs:\n");
  for (i = 0; i < privn; i++) {
    send_to_char(c, "%s%s^n ", privs[i].col, privs[i].privname);
  }
  send_to_char(c, "\n");
}

void command_grant(character *c, char *r) 
{
  if (STREQ(r, "")) {
    command_grant_usage(c);
    return;
  }
  {
    character *target;
    char *privname = r;
    const priv *p;
   
    strmunch(r, &r, &privname);
    /* r should now be player name, privname should be the priv */
    /* if privname is blank print out usage and bail out */
    if (STREQ(privname, "")) {
      command_grant_usage(c);
      return;
    }
    /* try to find player, if not found bail out of function */
    target = character_from_name(r);
    if (!target) {
      send_to_char(c, "Player %s not found\n", r);
      command_grant_usage(c);
      return;
    }
    /* search through the priv list and try to find the priv */
    p = privname_to_priv(privname);
    if (p) {
      send_to_char(c, "You have granted %s the %s priv.\n",
                   target->name, p->privname);
      send_to_char(target, "You have been granted the %s priv\n",
                   p->privname);
      setpriv(target, p->priv);
    } else {
      send_to_char(c, "Priv %s not found\n", privname);
      command_grant_usage(c);
    }
  }
}

void command_lsd(character *c, char *r)
{
  listnode *dNode = AllConns.head.next;
  descriptor *des2;
  char buff[OUTPUT_BUFFER];
  unsigned long howlong;
  time_t now;
  
  des2 = getdes(c);

  bar2(des2, "Sockets connected and states", "^Y", "^B");
  send_to_char(c, "^YSFD State    (name)        (login)  Idle    Host^n\n\n");
  while (LIST_NODE_IS_REAL(dNode)) {
    descriptor *des = LIST_GET_DATA(dNode, descriptor *, descriptorlink);
    character *c2 = &des->player;
    send_to_char(c, "%.3d ", des->sfd);
    switch(des->state) {
      case STATE_LOGIN :
        strcpy(buff, "LOGIN");
        break;
      case STATE_NEW1 :
        strcpy(buff, "NEW1");
        break;
      case STATE_NEW2 :
        strcpy(buff, "NEW2");
        break;
      case STATE_NEW3 :
        strcpy(buff, "NEW3");
        break;
      case STATE_PASSWORD :
        strcpy(buff, "PASSWORD");
        break;
      case STATE_PLAY :
        strcpy(buff, "PLAY");
        break;
      case STATE_CLOSING :
        strcpy(buff, "CLOSING");
        break;
    }
    send_to_char(c, "%-9s", buff);
    now = time(NULL);
    if(des->state == STATE_PLAY) {
      send_to_char(c, "%-13s", c2->name);
      /* login Time */
      howlong = (unsigned long) difftime(now, c2->logintime);
      send_to_char(c, " %02lu:", howlong / 3600);
      howlong %= 3600;
      send_to_char(c, "%02lu:", howlong / 60);
      howlong %= 60;
      send_to_char(c, "%02lu", howlong);
    } else {
      send_to_char(c, "                      ");
    }
    /* Idle Time */
    howlong = (unsigned long) difftime(now, des->idletime);
    send_to_char(c, " %02lu:", howlong / 60);
    howlong %= 60;
    send_to_char(c, "%02lu   %s\n", howlong, des->hostname);
    dNode = dNode->next;
  }
  bar2(des2, NULL, NULL, "^B");
}

void command_bug(character *c, char *r)
{
  if (STREQ(r, "")) {
    help_usage(c, "bug");
  } else {
    log(bug, "%s - %s", c->name, r);
    send_to_char(c, "Thank you, the bug has been recorded\n");
  }
}

void command_idea(character *c, char *r)
{
  if (STREQ(r, "")) {
    help_usage(c, "idea");
  } else {
    log(idea, "%s - %s", c->name, r);
    send_to_char(c, "Thank you, your idea has been recorded\n");
  }
}

void command_typo(character *c, char *r)
{
  if (STREQ(r, "")) {
    help_usage(c, "typo");
  } else {
    log(typo, "%s - %s", c->name, r);
    send_to_char(c, "Thank you, the typo has been recorded\n");
  }
}

void command_shutdown(character *c, char *r)
{
  sdown = TRUE;
}

void command_motd(character *c, char *r)
{
  descriptor *des = getdes(c);

  bar2(des, "Message Of The Day", NULL, NULL);
  send_file_to_descriptor("lib/etc/motd", des);
  bar2(des, NULL, NULL, NULL);
}

void command_prompt(character *c, char *r)
{
  if(strlen(r) > MAX_PROMPT_LENGTH-2) {
    send_to_char(c, "Prompt too long (max %d)\n", MAX_PROMPT_LENGTH-2);
  } else {
    strcpy(c->prompt, r);
/*    strcat(c->prompt, "^n");*/
  } 
}

void command_commands(character *c, char *r)
{
  descriptor *des = getdes(c);
  int i;

  bar2(des, "List of commands", NULL, NULL);
  for(i = 0;i < commandn;i++) {
    if(haspriv(c, commands[i].priv))
      send_to_char(c, "%s%s^n ", privcol(commands[i].priv), commands[i].name);
  }
  send_to_char(c, "\n");
  bar2(des, NULL, NULL, NULL);
}

void command_users(character *c, char *r)
{
  listnode *pNode = AllPlayers.head.next;

  while (LIST_NODE_IS_REAL(pNode)) {
    character *pChar = LIST_GET_DATA(pNode,character *, characterlink);
    send_to_char(c, "%s ", pChar->name);
    pNode = pNode->next;
  }
  send_to_char(c, "\n");
}

void command_who(character *c, char *r)
{
  descriptor *des = getdes(c);

  listnode *pNode = AllPlayers.head.next;
  int howmany = 0;
  time_t now;
  unsigned long howlong;

  now = time(NULL);
  bar2(des, "Who is logged on?", "^Y", "^G");
  send_to_char(c, "^Y%-*s Login    Idle    %-*s ",
               MAX_NAME_LENGTH, "Name", MAX_GROUP_NAME_LENGTH, "Group");
  if(haspriv(c, PRIV_SU_TRACE)) {
    send_to_char(c, "Host");
  }
  send_to_char(c, "^n\n");
  while (LIST_NODE_IS_REAL(pNode)) {
    character *pChar = LIST_GET_DATA(pNode, character *, characterlink);
    descriptor *des = LIST_GET_DATA(pChar, descriptor *, player);

    /* Display name */
    send_to_char(c, "%-*s", MAX_NAME_LENGTH, pChar->name);
    /* Calculate how long they have been logged in */
    howlong = (unsigned long) difftime(now, pChar->logintime);
    send_to_char(c, " %02lu:", howlong / 3600);
    howlong %= 3600;
    send_to_char(c, "%02lu:", howlong / 60);
    howlong %= 60;
    send_to_char(c, "%02lu", howlong);
    /* Calculate how long they have been idle for */
    howlong = (unsigned long) difftime(now, des->idletime);
    send_to_char(c, " %02lu:", howlong / 60);
    howlong %= 60;
    send_to_char(c, "%02lu   ", howlong);
    send_to_char(c, "%-*s ", MAX_GROUP_NAME_LENGTH, pChar->group);
    if(haspriv(c, PRIV_SU_TRACE)) {
      send_to_char(c, "%s", des->hostname);
    }
    send_to_char(c, "\n");
    pNode = pNode->next;
    howmany++;
  }
  send_to_char(c, "Total: %d user%s\n", howmany, howmany != 1 ? "s" : "");
  bar2(des, NULL, NULL, "^G");
}

void command_tell(character *c, char *sRemainder)
{
  char *sMessage = sRemainder;
  char *sTerm;

  while (*sMessage && !isspace(*sMessage))
    sMessage++;
  sTerm = sMessage;
  while (*sMessage && isspace(*sMessage))
    sMessage++;
  *sTerm = 0;

  {
    character *pDest = character_from_name(sRemainder);

    if (!pDest) {
      send_to_char(c, "Character '%s' could not be found or is ambiguous.\n", sRemainder);
      return;
    }

    if (pDest == c) {
      send_to_char(c, "Talking to yourself again?\n");
      return;
    }
    
    send_to_char(pDest, "^W> %s tells you '%s^W'^n\n", c->name, sMessage);
    send_to_char(c, "You tell %s '%s^n'\n", pDest->name, sMessage);
  }
}

void command_remote(character *c, char *sRemainder)
{
  char *sMessage = sRemainder;
  char *sTerm;

  while (*sMessage && !isspace(*sMessage))
    sMessage++;
  sTerm = sMessage;
  while (*sMessage && isspace(*sMessage))
    sMessage++;
  *sTerm = 0;

  {
    character *pDest = character_from_name(sRemainder);

    if (!pDest) {
      send_to_char(c, "Character '%s' could not be found or is ambiguous.\n", sRemainder);
      return;
    }

    if (pDest == c) {
      send_to_char(c, "Thats you, you fool\n");
      return;
    }
    
    send_to_char(pDest, "^W> %s %s^W, at you^n\n", c->name, sMessage);
    send_to_char(c, "You emote '%s %s^n' to %s^n\n", c->name, sMessage,
                 pDest->name);
  }
}

void command_say(character *c, char *r)
{
  if (STREQ(r, "")) {
    help_usage(c, "say");
    return;
  }
  send_to_room_except(c->group, c, "%s says \'%s^n\'\n", c->name, r);
  send_to_char(c, "You say \'%s^n\'\n", r);
}

void command_emote(character *c, char *r)
{
  if (STREQ(r, "")) {
    help_usage(c, "emote");
    return;
  }
  send_to_room_except(c->group, c, "%s %s\n", c->name, r);
  send_to_char(c, "You emote \'%s %s\'\n", c->name, r);
}

void command_shout(character *c, char *r)
{
  if (STREQ(r, "")) {
    help_usage(c, "shout");
    return;
  }
  send_to_all_except(c, "%s shouts \'%s^n\'\n", c->name, r);
  send_to_char(c, "You shout \'%s^n\'\n", r);
}

void command_quit(character *c, char *r)
{
  descriptor *des = getdes(c); 

  des->state = STATE_CLOSING;
  send_to_all_except(c, "\n+++ ^r%s^n has logged out\n", c->name);
  send_to_char(c, "\n+++ You have logged out\n");
  if (send_file_to_descriptor("lib/etc/logout", des) !=0)
    send_to_descriptor(des, "Goodbye!\n");

  des->safe = FALSE;
}

void command_whoami(character *c, char *r)
{
  send_to_char(c, "You are ^r%s^n\n", c->name);
}

bool command_shortcut_run(descriptor *des, char *s)
{
  character *p = &des->player;
  int c;

  s = strip_leading_white_space(s);
  if (strlen(s) != 0) {
    for (c = 0; c < commandn; c++) {
      if (strchr(commands[c].shortcut, s[0]) && haspriv(p, commands[c].priv)) {
        void (*function)(character *, char *);

        function = commands[c].function;
        s++; /* to get past the shortcut character */
        s = strip_leading_white_space(s);
        function(p, s);
        return TRUE;
      }
    }
  }

  return FALSE;
}

void command_parse_n_run(descriptor *des, const char *first,
                         char *remainder)
{
  unsigned int count = 0;
  unsigned int index = 0;
  unsigned int c;
  character *p = &des->player;

  if (STREQ(first, ""))
    return;

  for (c = 0; c < commandn; c++) {
    if (STREQ(first, commands[c].name) && haspriv(p, commands[c].priv)) {
      /* Matches exactly */
      void (*function)(character *, char *);

      function = commands[c].function;
      function(p, remainder);
      return;
    } else if (STRNEQ(first, commands[c].name, strlen(first))
               && haspriv(p, commands[c].priv)) {
      /* Part matches */
      index = c;
      count ++;
    }
  }
  switch (count) {
    case 0:
      send_to_char(p, "Command \'%s\' not known, try help\n", first);
      break;
    case 1:
      /* Matches one command without ambiguity */
      {
        void (*function)(character *, char *);

        function = commands[index].function;
        function(p, remainder);
        return;
        break;
      }
    default:
      /* Part matches more than one command */
      send_to_char(p, "Command \'%s\' ambiguous, try help\n", first);
      break;
   }
}

void command_do(descriptor *des, const char *buffer)
{
  char line[MAX_RAW_INPUT_BUFFER];
  char *first, *remainder;
  char *temp;
  bool carryon = TRUE;

  strcpy(line, buffer);
  first = line;

  while (carryon == TRUE) {
    /* strip leading white space in command */
    first = strip_leading_white_space(first);
    /* search for a ; in the input */
    temp = strchr(first+1, ';');
    if (temp != NULL) {
      /* found a ; */
      *temp = '\0';
      temp++;
      if (*temp == '\0') {
        /* is temp a 0 length string ? */
        carryon = FALSE;
      } else {
        carryon = TRUE;
      }
    } else {
      carryon = FALSE;
    }

    strip_trailing_white_space(first);  
    if (command_shortcut_run(des, first) == FALSE) {
      strmunch(first, &first, &remainder);
      strlower(first);
      command_parse_n_run(des, first, remainder);
    }

    if (carryon == TRUE) {
      first = temp;
    }
  }

  /* Do a prompt */
  send_to_char(&des->player, "%s^n", &des->player.prompt);
}

descriptor* getdes(character *c) {
  return LIST_GET_DATA(c, descriptor *, player);
}

void seconds_to_string(unsigned long seconds, char *buffer)
{
  unsigned long t;

  t = seconds / (60 * 60 * 24);
  if (t > 0)
    buffer += sprintf(buffer, "%lu day%s, ", t, t != 1 ? "s" : "");
  seconds %= (60 * 60 * 24);
  t = seconds / (60 * 60);
  if (t > 0)
    buffer += sprintf(buffer, "%lu hr%s, ", t, t != 1 ? "s" : "");
  seconds %= (60 * 60);
  t = seconds / 60;
  if (t > 0)
    buffer += sprintf(buffer, "%lu min%s, ", t, t != 1 ? "s" : "");
  seconds %= 60;
  sprintf(buffer, "%lu sec%s", seconds, seconds != 1 ? "s": "");
}
